#include "stdafx.h"
#include "Insomnia.h"

// Manifest pulls in the right version of ComCtl32.dll
#pragma comment(linker, "\"/manifestdependency:type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")

// Constants
#define WINDOW_BACKGROUND (RGB(220, 220, 220))
#define WM_NOTIFYICON (WM_USER + 10)
#define MAX_LOADSTRING (80)

// Global variables
TCHAR szTitle[MAX_LOADSTRING] = {0};
TCHAR szText[MAX_LOADSTRING] = {0};
TCHAR szVersion[MAX_LOADSTRING] = {0};
TCHAR szLink[MAX_LOADSTRING] = {0};
TCHAR szStesError[MAX_LOADSTRING] = {0};
HINSTANCE g_hInstance = NULL;
HWND g_hWnd = NULL;
HWND hWndText = NULL;
HWND hWndVersion = NULL;
HWND hWndLink = NULL;
HWND hWndStatusArea = NULL;

LRESULT CALLBACK WndProc(const HWND hWnd, const UINT message, const WPARAM wParam, const LPARAM lParam)
{
	static BOOL fNotifyBalloonShown = FALSE;

	switch (message)
	{
	case WM_CREATE:
		{
			// Center window on the desktop
			const HWND hWndDesktop = GetDesktopWindow();
			RECT ownerRect = {0};
			if (GetWindowRect(hWndDesktop, &ownerRect))
			{
				const CREATESTRUCT* const pCreateStruct = (CREATESTRUCT*)lParam;
				SetWindowPos(hWnd, NULL, ((ownerRect.right - ownerRect.left) - pCreateStruct->cx) / 2, ((ownerRect.bottom - ownerRect.top) - pCreateStruct->cy) / 2, 0, 0, SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
			}
		}
		return 0;
	case WM_SIZE:
		{
			NOTIFYICONDATA notifyIconData = {0};
			notifyIconData.cbSize = sizeof(notifyIconData);
			notifyIconData.uFlags = NIF_MESSAGE;
			notifyIconData.hWnd = hWnd;
			notifyIconData.uCallbackMessage = WM_NOTIFYICON;
			switch(wParam)
			{
			case SIZE_RESTORED:
				{
					// Layout controls
					const LONG margin = 4;
					const WORD availableWidth = LOWORD(lParam);
					WORD availableHeight = HIWORD(lParam);
					SIZE idealVersionSize = {0};
					(void)SendMessage(hWndVersion, LM_GETIDEALSIZE, availableWidth, (LPARAM)&idealVersionSize);
					MoveWindow(hWndVersion, availableWidth - idealVersionSize.cx - margin, availableHeight - idealVersionSize.cy - margin, idealVersionSize.cx, idealVersionSize.cy, TRUE);
					SIZE idealLinkSize = {0};
					(void)SendMessage(hWndLink, LM_GETIDEALSIZE, availableWidth, (LPARAM)&idealLinkSize);
					MoveWindow(hWndLink, margin, availableHeight - idealLinkSize.cy - margin, idealLinkSize.cx, idealLinkSize.cy, TRUE);
					const LONG statusHeight = ((idealVersionSize.cy <= idealLinkSize.cy) ? idealLinkSize.cy : idealVersionSize.cy) + margin + margin;
					availableHeight -= (WORD)statusHeight;
					MoveWindow(hWndStatusArea, 0, availableHeight, availableWidth, statusHeight, TRUE);
					SIZE idealTextSize = {0};
					(void)SendMessage(hWndText, LM_GETIDEALSIZE, availableWidth, (LPARAM)&idealTextSize);
					MoveWindow(hWndText, (availableWidth - idealTextSize.cx) / 2, (availableHeight - idealTextSize.cy) / 2, idealTextSize.cx, idealTextSize.cy, TRUE);

					// Remove notification icon (if present)
					(void)Shell_NotifyIcon(NIM_DELETE, &notifyIconData);
				}
				break;
			case SIZE_MINIMIZED:
				// Add notification icon
				if (SUCCEEDED(LoadIconMetric(g_hInstance, MAKEINTRESOURCE(IDI_INSOMNIA), LIM_SMALL, &(notifyIconData.hIcon))))
				{
					notifyIconData.uFlags |= NIF_ICON;
				}
				if (SUCCEEDED(StringCchCopy(notifyIconData.szTip, ARRAYSIZE(notifyIconData.szTip), szTitle)))
				{
					notifyIconData.uFlags |= NIF_TIP;
				}
				if (!fNotifyBalloonShown && SUCCEEDED(StringCchCopy(notifyIconData.szInfo, ARRAYSIZE(notifyIconData.szInfo), szTitle)))
				{
					notifyIconData.uFlags |= NIF_INFO;
					fNotifyBalloonShown = true;
				}
				(void)Shell_NotifyIcon(NIM_ADD, &notifyIconData);

				// Remove taskbar button
				(void)ShowWindow(hWnd, SW_HIDE);
				break;
			}
		}
		return 0;
	case WM_CTLCOLORSTATIC:
		if ((HWND)lParam == hWndText)
		{
			// Change color of text message to red
			const HDC hdc = (HDC)wParam;
			(void)SetTextColor(hdc, RGB(255, 0, 0));
			(void)SetBkMode(hdc, WINDOW_BACKGROUND);
			return (INT_PTR)CreateSolidBrush(WINDOW_BACKGROUND);
		}
		break;
	case WM_NOTIFY:
		{
			const NMHDR* const pnmHdr = (NMHDR*)lParam;
			if ((pnmHdr->hwndFrom == hWndLink) &&
				((NM_CLICK == pnmHdr->code) || (NM_RETURN == pnmHdr->code)))
			{
				// Handle click (or enter) on URL by opening it
				const NMLINK* const pnmLink = (NMLINK*)pnmHdr;
				SHELLEXECUTEINFO sei = {0};
				sei.cbSize = sizeof(sei);
				sei.hwnd = hWnd;
				sei.lpFile = pnmLink->item.szUrl;
				sei.nShow = SW_NORMAL;
				(void)ShellExecuteEx(&sei);
			}
		}
		return 0;
	case WM_NOTIFYICON:
		{
			const WORD msg = LOWORD(lParam);
			if ((WM_LBUTTONDOWN == msg) || (WM_RBUTTONDOWN == msg))
			{
				// Handle left/right click on notify icon by restoring the window
				(void)ShowWindow(hWnd, SW_NORMAL);
			}
		}
		return 0;
	case WM_DESTROY:
		// Exit the application
		(void)PostQuitMessage(0);
		return 0;
	}

	// Default processing
	return DefWindowProc(hWnd, message, wParam, lParam);
}

BOOL InitInstance(const HINSTANCE hInstance, const int nCmdShow)
{
	// Create application window
	const HWND hWnd = CreateWindowEx(WS_EX_TOPMOST, szTitle, szTitle, WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX, CW_USEDEFAULT, CW_USEDEFAULT, 300, 200, NULL, NULL, hInstance, NULL);
	if (NULL == hWnd)
	{
		return FALSE;
	}
	g_hWnd = hWnd;

	// Create child windows
	hWndText = CreateWindowEx(0, WC_LINK, szText, WS_CHILD | WS_VISIBLE, 0, 0, 0, 0, hWnd, NULL, hInstance, NULL);
	hWndStatusArea = CreateWindowEx(0, WC_STATIC, NULL, WS_CHILD | WS_VISIBLE | SS_SUNKEN, 0, 0, 0, 0, hWnd, NULL, hInstance, NULL);
	hWndVersion = CreateWindowEx(0, WC_LINK, szVersion, WS_CHILD | WS_VISIBLE, 0, 0, 0, 0, hWnd, NULL, hInstance, NULL);
	hWndLink = CreateWindowEx(0, WC_LINK, szLink, WS_CHILD | WS_VISIBLE | WS_TABSTOP, 0, 0, 0, 0, hWnd, NULL, hInstance, NULL);
	if ((NULL == hWndText) || (NULL == hWndStatusArea) || (NULL == hWndVersion) || (NULL == hWndLink))
	{
		return FALSE;
	}

	// Customize control fonts
	const HFONT hFontText = CreateFont(20, 0, 0, 0, FW_BOLD, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_SWISS, NULL);
	(void)SendMessage(hWndText, WM_SETFONT, (WPARAM)hFontText, FALSE);
	const HFONT hFontVersionLink = CreateFont(15, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_SWISS, TEXT("Calibri"));
	(void)SendMessage(hWndVersion, WM_SETFONT, (WPARAM)hFontVersionLink, FALSE);
	(void)SendMessage(hWndLink, WM_SETFONT, (WPARAM)hFontVersionLink, FALSE);

	// Show application window
	(void)ShowWindow(hWnd, nCmdShow);
	(void)UpdateWindow(hWnd);
	return TRUE;
}

int APIENTRY _tWinMain(const HINSTANCE hInstance, const HINSTANCE hPrevInstance, const LPTSTR lpCmdLine, const int nCmdShow)
{
	UNREFERENCED_PARAMETER(hPrevInstance);
	UNREFERENCED_PARAMETER(lpCmdLine);
	g_hInstance = hInstance;

	//Load resource strings
	if ((0 == LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING)) ||
		(0 == LoadString(hInstance, IDS_APP_TEXT, szText, MAX_LOADSTRING)) ||
		(0 == LoadString(hInstance, IDS_APP_VERSION, szVersion, MAX_LOADSTRING)) ||
		(0 == LoadString(hInstance, IDS_APP_LINK, szLink, MAX_LOADSTRING)) ||
		(0 == LoadString(hInstance, IDS_STES_ERROR, szStesError, MAX_LOADSTRING)))
	{
		return FALSE;
	}

	// SysLink class info
	INITCOMMONCONTROLSEX icsex = {0};
	icsex.dwSize = sizeof(icsex);
	icsex.dwICC = ICC_LINK_CLASS;

	// Window class info
	WNDCLASSEX wcex = {0};
	wcex.cbSize = sizeof(wcex);
	wcex.lpfnWndProc = WndProc;
	wcex.hInstance = hInstance;
	wcex.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_INSOMNIA));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = CreateSolidBrush(WINDOW_BACKGROUND);
	wcex.lpszClassName = szTitle;

	// Initialize
	if (!InitCommonControlsEx(&icsex) ||
		(0 == RegisterClassEx(&wcex)) ||
		!InitInstance(hInstance, nCmdShow))
	{
		return FALSE;
	}

	MSG msg = {0};
	// Prevent the system from entering sleep mode
	const EXECUTION_STATE previousExecutionState = SetThreadExecutionState(ES_CONTINUOUS | ES_SYSTEM_REQUIRED);
	if (NULL != previousExecutionState)
	{
		// Message loop
		while (GetMessage(&msg, NULL, 0, 0))
		{
			if (!IsDialogMessage(g_hWnd, &msg))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}

		// Restore previous state
		(void)SetThreadExecutionState(previousExecutionState);
	}
	else
	{
		// Display error
		(void)MessageBox(g_hWnd, szStesError, szTitle, MB_OK | MB_ICONERROR);
	}

	return (int)msg.wParam;
}
